(* moodul Dijkstra teostab sammsammulist Dijkstra algoritmi lbimngu *) 

open Struktuurid;;
open AlgoBaas;;

let kaugused = Hashtbl.create 10;; 		(* sisuliselt map tipp.nimi : int *)

let valitudTipp = ref(tyhiTipp);;			(* valitud tipp *)

(* funktsioon, mis tagastab kahe tipu puhul -1, kui 1. kaugus on viksem, 1, kui 2. kaugus on viksem, 0 muudel juhtudel *)
let v2iksemaKaugusega tipp1 tipp2 =
	let k1 = Hashtbl.find kaugused tipp1.nimi in
	let k2 = Hashtbl.find kaugused tipp2.nimi in
	if k1 < k2 then -1 else if k1 > k2 then 1 else 0;; 

(* funktsioon, mis sorteerib tippude listi kauguste phjal algtipust *)
let sordiTipud(tipud) =
	List.sort v2iksemaKaugusega tipud;;

(*funktsioon, mis tagastab snena kikide mittevaadeldud tippude kaugused algtipust *)
let string_of_mitteVaadelduteKaugused(tipud) =
	(*let vaatlemataTipud = List.filter (fun t -> !(t.tv) <> Vaadeldud) tipud in	(* mittevaadeldud tipud *)
	let s = List.fold_left (fun a b -> a ^ ", " ^ b.nimi ^ ": " ^ (if Hashtbl.find kaugused b.nimi = max_int then "inf" else string_of_int(Hashtbl.find kaugused b.nimi))) "" vaatlemataTipud in
	"Ttlemata tippude kaugused algtipust: [" ^ (if s = "" then "" else String.sub s 2 (String.length s - 2)) ^ "]";;*)
	let vaadeldavadTipud = sordiTipud(List.filter (fun t -> !(t.tv) = Vaadeldav) tipud) in	(* mittevaadeldud tipud *)
	let s = List.fold_left (fun a b -> a ^ ", " ^ b.nimi ^ ": " ^ (if Hashtbl.find kaugused b.nimi = max_int then "inf" else string_of_int(Hashtbl.find kaugused b.nimi))) "" vaadeldavadTipud in
	"Tippude kauguste eelistusjrjekord: [" ^ (if s = "" then "" else String.sub s 2 (String.length s - 2)) ^ "]";;

(*funktsioon, mis tagastab snena kikide tippude kaugused algtipust *)
let string_of_kaugused() =
	let s = Hashtbl.fold (fun k v acc -> k ^ ": " ^ (if v = max_int then "inf" else string_of_int(v)) ^ ", " ^ acc) kaugused "" in
  "Tippude kaugused algtipust: " ^ String.sub s 0 (String.length(s) - 2);;

(* funktsioon, mis leiab kik tipust tipp vljuvad servad *)
let leiaV2ljuvadServad(tipp, servad) =
	List.filter ((fun t s -> !(s.tipp1) = t) tipp) servad;;

(* funktsioon, mis mrab tipule kauguse *)
let lisaKaugus kaugus tipp =
	Hashtbl.replace kaugused tipp.nimi kaugus;;

(* funktsioon, mis muudab tipust tipp vljuva serva teises otsas oleva tipu kaugust (tipu tipp kaudu), *)
(* kui see on senisest lhem*)
let uuendaKaugust(serv) =
	match serv with
		| {tipp1 = t1; tipp2 = t2; kaal = Some k;} -> (
			let senineKaugus = Hashtbl.find kaugused !t2.nimi in
			let uusKaugus = (Hashtbl.find kaugused !t1.nimi) + k in
			if uusKaugus < senineKaugus 
				then (
					lisaKaugus uusKaugus !t2;
					(*!t2.hind := Some uusKaugus*)
				)
		)
		| _ -> failwith("Serval puudub kaal. Seda et tohiks juhtuda.");;

(* funktsioon, mis mrgib serva ja tema tipp2 vaadeldavaks (v.a. kui viimane juba valitud/vaadeldud on) *)
let vaatleServa(serv) =
	serv.sv := Vaadeldav;
	if !(!(serv.tipp2).tv) = Vaatlemata then !(serv.tipp2).tv := Vaadeldav;;

(* funktsioon, mis mrgib serva ja tema tipp2 valituks (v.a. kui viimane juba vaadeldud on) *)
let valiServ(serv) =
	serv.sv := Valitud;
	if !(!(serv.tipp2).tv) <> Vaadeldud then !(serv.tipp2).tv := Valitud;;

(* funktsioon, mis mrgib serva ja tema tipp2 vaadelduks *)
let lisaServ(serv) =
	serv.sv := Vaadeldud;
	!(serv.tipp2).tv := Vaadeldud;;

(* funktsioon, mis tagastab, kas serv on valitav, st on vaadeldav ja tema tipp2 pole vaadeldud *)
let valitavServ(serv) =
	match serv with
		| {tipp2 = t2; sv = v} -> (
			!v = Vaadeldav && !(!t2.tv) <> Vaadeldud
		);;

(* funktsioon, mis tagastab tipu, mille kaugus algtipust on kige viksem *)
let leiaL2himTipp(tipud) =
	List.fold_left (fun t1 t2 -> if Hashtbl.find kaugused t1.nimi <=  Hashtbl.find kaugused t2.nimi then t1 else t2) (List.hd tipud) (List.tl tipud);;

(* funktsioon, mis tagastab, kas serv on tipule vastav, st mille teise tipu kaugus + serv.kaal on tipu kaugus *)
let vastavServ tipp serv =
	match serv.kaal with
		| None -> false
		| Some k -> !(serv.tipp2) = tipp && Hashtbl.find kaugused !(serv.tipp1).nimi + k = Hashtbl.find kaugused tipp.nimi;;

(* algoritmi algus, mille kigus mratakse kikidele tippudele kaugused algtipust *)
let algus(algtipp, tipud, servad) =
	List.iter (lisaKaugus max_int) tipud; 			(* paneme kikidele tippudele algseks kauguseks algtipust suurima vimaliku *)
	lisaKaugus 0 algtipp; 											(* ainult algtipule paneme kauguseks 0 *)
	tekst := "Dijkstra algoritm alustab valitud algtipust. Mrame kikidele tippudele kaugused algtipust: algtipule 0, kikidele teistele lpmatuse.";
	(*List.iter (fun t -> t.hind := Some (Hashtbl.find kaugused t.nimi)) tipud;*)
	nk1 := string_of_mitteVaadelduteKaugused(tipud);
	valitudTipp := algtipp;
	i := ServaLisamine;;

(* valitud tipu ja vastava serva vaadelduks mrkimine *)
let servaLisamine(algtipp, tipud, servad) =
	if !valitudTipp = algtipp																								(* kui valitud tipp on algtipp *)
		then algtipp.tv := Vaadeldud																					(* mrgime ta vaadelduks *)
	else List.iter (fun s -> if !(s.sv) = Valitud then lisaServ(s)) servad; (* muidu mrgime serva koos tipuga vaadelduks*)
	!valitudTipp.hind := Some (Hashtbl.find kaugused !valitudTipp.nimi);
	tekst := "Mrgime valitud tipu vaadelduks.";
	nk1 := string_of_mitteVaadelduteKaugused(tipud);
	if List.for_all (fun t -> !(t.tv) = Vaadeldud) tipud										(* kui kik tipu on vaadeldud, lhme lpule *)
  		then i := Lopp
	else i := ServaVaatlus;;																								(* vastasel juhul lhme tippude kaugusi uuendama *)

(* sja vaadeldud tipust vljuvate servade leidmine, vaadeldavateks mrkimine ja nende sihttippude kauguste uuendamine *)
let servaVaatlus(tipud, servad) =
  	let vaadeldavadServad = leiaV2ljuvadServad(!valitudTipp, servad) in		(* leiame sja klastatud tipust vljuvad servad *)
  	List.iter vaatleServa vaadeldavadServad;															(* mrgime need vaadeldavateks *)
  	List.iter uuendaKaugust vaadeldavadServad;														(* uuendame nende 2. tippude kaugusi algtipust *)
  	tekst := "Vaatleme sja vaadeldud tipust vljuvaid servi ning uuendame servade sihttippude kaugusi, kui need on senisest viksemad.";
		nk1 := string_of_mitteVaadelduteKaugused(tipud);
  	i := ServaValik;;

(* algtipule lhima klastamata tipu ja vastava serva leidmine ja valituks mrkimine *)
let servaValik(tipud, servad) =
	let t = leiaL2himTipp(List.filter (fun t -> !(t.tv) = Vaadeldav) tipud) in		(* leiame algtipule lhima klastamata tipu*)
	let s = List.find (vastavServ t) servad in																		(* ja talle vastava serva *)
	valiServ(s);																																	(* mrgime serva ja tipu valituks *)
	valitudTipp := t;
	tekst := "Valime algtipule lhima klastamata tipu.";
	nk1 := string_of_mitteVaadelduteKaugused(tipud);
	i := ServaLisamine;;

(* algoritmi lpp *)
let lopp(tipud) =
	tekst := "Eelistusjrjekord sai thjaks. Algoritm lpetab, olles leidnud kikide tippude vhima kauguse algtipust. Pildil on tekkinud kauguste puu.";
	nk1 := string_of_kaugused();
	AlgoBaas.lopp();;

(* algoritmi samm *)
let samm(algtipp, tipud, servad) =
	match !i with
		| Algus -> algus(algtipp, tipud, servad)
		| ServaVaatlus -> servaVaatlus(tipud, servad)
		| ServaLisamine -> servaLisamine(algtipp, tipud, servad)
		| ServaValik -> servaValik(tipud, servad)
		| Lopp -> lopp(tipud)
		| _ -> ();;