(* moodul Kosaraju teostab sammsammulist Kosaraju algoritmi lbimngu *) 

open Struktuurid;;
open AlgoBaas;;

let sygavutiTipud = ref([]);;
let komponendid = ref([]);; 		(* kik sidusad komponendid. TODO: hoopis hulgana teha? *)
let komponent = ref([]);; 			(* ks sidus komponent *)

(* pran servad vastupidi, st vahetan tipud ra *)
let rec ps vanadAndmed serv =
	match serv with
		| {tipp1 = t1; tipp2 = t2;} -> (
			let (vanaX, vanaY, vanaR, vanaK) = Hashtbl.find vanadAndmed (!t1.nimi ^ ":" ^ !t2.nimi) in
			let t = !t1 in 
			serv.tipp1 := !t2;
			serv.tipp2 := t; 
			(*muudaKaareAndmeid(!(x.tipp1), !(x.tipp2));*)
			uuendaKaareAndmeid(!(serv.tipp1), !(serv.tipp2), vanaX, vanaY, vanaR, vanaK)
		);;

let pooraServad(servad) =
	let vanadAndmed = Hashtbl.create 10 in
	List.iter (fun s -> let n = !(s.tipp1).nimi ^ ":" ^ !(s.tipp2).nimi in Hashtbl.add vanadAndmed n 
			(Hashtbl.find kaareX n, Hashtbl.find kaareY n, Hashtbl.find kaareR n, Hashtbl.find kaareX n)) servad;
	List.iter (ps vanadAndmed) servad;;


(* funktsioon sgavuti lppjrjestuse snena esitamiseks *)
let string_of_lopp(tipud) =
	"Lppjrjestus: "  ^ string_of_tipud(tipud);;

(* funktsioon tagurpidi sgavuti lppjrjestuse snena esitamiseks *)
let string_of_tagurpidi(tipud) =
	"Tagurpidi lppjrjestus: "  ^ string_of_tipud(tipud);;		(* NB! mitte List.rev, selle prasin juba ringi *)
	
(* funktsioon he tugevalt sidusa komponendi snena esitamiseks *)
let string_of_komponent(tipud) =
	"Tugevalt sidus komponent: " ^ string_of_tipud(tipud);;

(* funktsioon tugevalt sidusate komponentide snena esitamiseks *)
let string_of_komponendid(komponendid) =
	"Tugevalt sidusad komponendid:" ^ List.fold_left  (fun a b -> a ^ " " ^ string_of_tipud b) "" komponendid;;

let algus(tipud, servad) =
	List.iter (fun t -> TopoKahn.uuendaSisendastet (TopoKahn.leiaSisendaste(t, servad)) t) tipud;
	tekst := "Kosaraju algoritm algustab valitud tipust.";
	i := Sygavuti;;

(* lbime graafi sgavuti lppjrjestuses. Kui mni tipp jb lbi kimata, siis mitu korda *)
let sygavuti(algtipp, tipud, servad) =
	sygavutiTipud := TopoLopp.l2biSygavuti(algtipp, tipud, servad);				(* lbime graafi sgavuti lppjrjestuses *)
	tekst := "Lbime graafi sgavuti lppjrjestuses ja kirjutame vlja tekkinud lppjrjestuse.";
	nk1 := string_of_lopp(!sygavutiTipud);
	i := PooratudGraaf;;

(* tekitame pratud kaartega graafi *)
let pooratudGraaf(tipud, servad) =
	tekst := "Tekitame pratud kaartega graafi. ";
	tekst := !tekst ^ "Prame tekkinud jrjestuse tagurpidi ja hakkame lbima tippe alates esimesest lbimata tipust tagurpidi jrjestuses.";
	sygavutiTipud := List.rev(!sygavutiTipud); (*tippude jrjestuse pidi ka mber prama*)
	nk2 := string_of_tagurpidi(!sygavutiTipud);
	pooraServad(servad);
	List.iter (fun s -> s.sv := Vaatlemata) servad;
	List.iter (fun t -> t.tv := Vaatlemata) tipud;
	i := EsimeneTipp;;

(* esimese komponenditu tipu valituks mrkimine *)
let esimeneTipp(tipud, servad) =
	List.iter (fun t -> if !(t.tv) = Vaadeldud then t.tv := Sobimatu) tipud;
	List.iter (fun s -> if !(s.sv) = Vaadeldud then s.sv := Sobimatu) servad;
	let esimeneTipp = List.find (fun t -> !(t.tv) = Vaatlemata) !sygavutiTipud in
	esimeneTipp.tv := Valitud;
	tekst := "Valime tagurpidi lppjrjestusest esimese tipu, mis pole veel heski komponendis.";
	nk1 := string_of_tagurpidi(!sygavutiTipud);
	nk2 := string_of_komponent(!komponent);
	nk3 := string_of_komponendid(!komponendid);
	i := TeisedTipud;;

(* valitud tipust alates graafi lbimine ja kikide lbitud tippude hendamine esimesega heks komponendiks *)
let teisedTipud(tipud, servad) =
	let esimeneTipp = List.find (fun t -> !(t.tv) = Valitud) tipud in
	i := Algus;
	while !algoL2bi = false
		do
			Laiuti.samm(esimeneTipp, tipud, servad)
		done;
	algoL2bi := false;
	List.iter (fun t -> if !(t.tv) = Valitud then t.tv := Vaadeldud) tipud;
	List.iter (fun s -> if !(s.sv) = Valitud then s.sv := Vaadeldud) servad;
	komponent := List.filter (fun t -> !(t.tv) = Vaadeldud) tipud;
	komponendid := !komponendid @ [!komponent];
	tekst := "Lisame temaga hte sidususkomponenti kik tipud, millesse saab temast mberpratud kaartega graafis juda.";
	nk1 := string_of_tagurpidi(!sygavutiTipud);
	nk2 := string_of_komponent(!komponent);
	nk3 := string_of_komponendid(!komponendid);
	if List.for_all (fun t -> !(t.tv) = Vaadeldud || !(t.tv) = Sobimatu) tipud
		then i := Lopp
	else i := EsimeneTipp;;

(* algoritmi lpp *)
let lopp(tipud, servad) =
	List.iter (fun t -> if !(t.tv) = Vaadeldud then t.tv := Sobimatu) tipud;
	List.iter (fun s -> if !(s.sv) = Vaadeldud then s.sv := Sobimatu) servad;
	tekst := "Kik tipud on vaadeldud. Kosaraju algoritm lpetab, olles leidnud graafi tugevalt sidusad komponendid.";
	nk1 := string_of_komponendid(!komponendid);
	nk2 := "";
	nk3 := ""; 
	AlgoBaas.lopp();;

(* algoritmi samm *)
let samm(algtipp, tipud, servad) = 
	match !i with
		| Algus -> algus(tipud, servad)
		| Sygavuti -> sygavuti(algtipp, tipud, servad)
		| PooratudGraaf -> pooratudGraaf(tipud, servad)
		| EsimeneTipp -> esimeneTipp(tipud, servad)
		| TeisedTipud -> teisedTipud(tipud, servad)
		| Lopp -> lopp(tipud, servad)
		| L2bi -> ()
		| _ -> ();;