(* moodul Kruskal teostab sammsammulist Kruskali algoritmi lbimngu *) 

open Struktuurid;;
open AlgoBaas;;

let hulgad = Hashtbl.create 10;; 	(* sisuliselt map tipp.nimi : tipp.nimi, tippude ja hulkade vastavuse hoiustamiseks *)

let sej = ref([]);; 							(* servade eelistusjrjekord *)

(* funktsioon, mis mrab igale t2 hulgaga tipule t1 hulga, et kigil sama hulk oleks*)
let lisaHulka(t1, t2) =
	let hulk1 = Hashtbl.find hulgad t1.nimi in
	let hulk2 = Hashtbl.find hulgad t2.nimi in
	Hashtbl.iter (fun k v -> if v = hulk2 then Hashtbl.replace hulgad k hulk1) hulgad;;

(* funkstioon, mis mrgib serva ja tema tipud valituks ning mrab, kas serva lisada vi mitte*)
let valiServ(serv) =
	match serv with
		| {tipp1 = t1; tipp2 = t2; sv = v;} -> (
			v := Valitud;
			tekst := "Valime vikseima kaaluga vaatlemata serva.";
			nk1 := string_of_sej(!sej);
			if Hashtbl.find hulgad !t1.nimi = Hashtbl.find hulgad !t2.nimi (*tekitabTsykli(!t1, !t2)*)
				then
					i := SobimatuServ
			else (
				lisaHulka(!t1, !t2);
				i := ServaLisamine
			);
			if !((!t1).tv) = Vaatlemata then (!t1).tv := Valitud;
			if !((!t2).tv) = Vaatlemata then (!t2).tv := Valitud
		);
		sej := List.tl !sej;;																									(* eemaldame serva eelistusjrjekorrast *)

(* funktsioon, mis muudab serva ja selle tippude vaadeldavust v1-st v2-ks *)
let muudaServa v1 v2 serv =
	match serv with
		| {tipp1 = t1; tipp2 = t2; sv = v;} -> (
			if !v = v1 then (
  			v := v2;
  			if !((!t1).tv) = v1 then (!t1).tv := v2;
  			if !((!t2).tv) = v1 then (!t2).tv := v2
			)
		);;

(* funktsioon, mid tagastab, kas tekkinud graaf on sidus (ehk kas kik tipud kuuluvad samasse hulka) *)
let onSidus(tipud) =
	let esimeseHulk = Hashtbl.find hulgad (List.hd tipud).nimi in
	List.for_all (fun t -> Hashtbl.find hulgad t.nimi = esimeseHulk) tipud;;

(* algoritmi algus *)
let algus() =
	tekst := "Kruskali algoritm alustab.";
	i := Vahe1;;

(* graafi servade eelistusjrjekorda lisamine *)
let vahe1(tipud, servad) =
	List.iter (fun t -> t.tv := Vaadeldud) tipud;						(* mrgime kik tipud vaadelduteks *)
	List.iter (fun t -> Hashtbl.add hulgad t.nimi t.nimi) tipud;		(* lisame igale tipule oma hulga *)
	List.iter (fun s -> sej := !sej @ [s]) servad;
	sej := sordiJ2rjekord(!sej);
	tekst := "Lisame kik servad eelistusjrjekorda. Koostatav graafi toes koosneb alguses ainult hetipulistest puudest.";
	nk1 := string_of_sej(!sej);
	if List.length !sej = 0
		then (
			List.iter (fun t -> Hashtbl.add hulgad t.nimi t.nimi) tipud;	(* lisame sellele hele tipule ka hulga *)  
			i := Lopp
		)
	else i := ServaValik;;

(* eelistusjrjekorrast lhima serva eemaldamine ja koos vastavate vaatlemata tippudega valituks mrkimine *)
let servaValik(servad) =
	match List.length !sej with																									(* servade eelistusjrjekorra pikkus *)
		| 0 -> failwith("htegi serva pole eelistusjrjekorras, nii ei tohiks juhtuda.")
		| _ -> valiServ(List.hd !sej);;																						(* lhim vaatlemata serv *)

(* kaht klastatud tippu hendava serva sobimatuks mrkimine ja eelistusjrjekorrast eemaldamine *)
let sobimatuServ(servad) =
	List.iter (fun s -> if !(s.sv) = Valitud then s.sv := Sobimatu) servad; 		(* mrgime serva sobimatuks *)
	tekst := "See serv hendab samas sidusas komponendis olevaid tippe, nii et seda me ei lisa. Eemaldame serva eelistusjrjekorrast.";
	nk1 := string_of_sej(!sej);
	if List.length !sej = 0
		then i := Lopp
	else i := ServaValik;;

(* valitud serva vaadelduks mrkimine ja eelistusjrjekorrast eemaldamine *)
let servaLisamine(servad, tipud) =
	let lisatavServ = List.find (fun s -> !(s.sv) = Valitud) servad in
	muudaServa Valitud Vaadeldud lisatavServ; 																	(* mrgime serva vaadelduks *)
	tekst := "See serv hendab eri sidusaid komponente olevaid tippe, nii et lisame selle ja eemaldame eelistusjrjekorrast.";
	nk1 := string_of_sej(!sej);
	if List.length !sej = 0
		then i := Lopp
	else i := ServaValik;;

(* algoritmi lpp *)
let lopp(tipud) =
	tekst := "Eelistusjrjekord on thi. Algoritm lpetab, olles leidnud minimaalse toes" ^ 
			(if onSidus(tipud) then "puu." else "metsa.");
	nk1 := string_of_sej(!sej);
	AlgoBaas.lopp();;

(* algoritmi samm *)
let samm(tipud, servad) = 
	match !i with
		| Algus -> algus()
		| Vahe1 -> vahe1(tipud, servad)
		| ServaValik -> servaValik(servad)
		| SobimatuServ -> sobimatuServ(servad)
		| ServaLisamine -> servaLisamine(servad, tipud)
		| Lopp -> lopp(tipud)
		| _ -> ();;