(* moodul Laiuti teostab sammsammulist graafi laiuti lbimist *) 

open Struktuurid;;
open AlgoBaas;;

let j2rgmisedServad = ref([]);;			(* jrgmised servad *)

let toodeldudTipud = ref([]);;			(* tdeldud tipud *)

(* funktsioon, mis tagastab, kas serv vljub tipust ning on vaatlemata *)
let sobivJ2rgServ tipp serv =
	match serv with 
		| {tipp1 = t1; tipp2 = t2; nool = n} ->
			match n with
				| false -> !t1 = tipp && !((!t2).tv) = Vaatlemata || !t2 = tipp && !((!t1).tv) = Vaatlemata	(* mittesuunatud graaf *)
				| true -> !t1 = tipp && !((!t2).tv) = Vaatlemata;;																					(* suunatud graaf *)

(* funktsioon, mis leiab tipu phjal jrgmised vaatlemata servad, mida klastada. Mittesuunatud graafide puhul sobivad servad, *)
(* mille esimene tipp on vastav tipp ja teine on vaatlemata vi vastupidi, suunatud graafide puhul sobivad vaid servad, mille*)
(* esimene tipp on vastav tipp ja teine on vaatlemata *)
let leiaJ2rgServad(tipp, servad) =
	List.filter (sobivJ2rgServ tipp) servad;;

(* leiab serva jrgi lisatava tipu (selle, mis on valitud) *)
let leiaLisatavTipp(serv) =
	match serv with
		| {tipp1 = t1; tipp2 = t2;} -> if !((!t1).tv) = Valitud then !t1 else !t2;;

(* funktsioon, mis eemaldab serva jrgmiste servade nimekirjast, juhul kui kumbki tema tipp pole enam vaatlemata *)
let rec eemalda(servad) =
	List.filter (fun s -> !(!(s.tipp1).tv) = Vaatlemata || !(!(s.tipp2).tv) = Vaatlemata) servad;;

(* funktsioon, mis jrgib serva ja vastava vaatlemata tipu valituks *)
let valiServ(serv) =
	match serv with
		| {tipp1 = t1; tipp2 = t2; sv = v;} -> (
			v := Valitud;
			if !((!t1).tv) = Vaatlemata then (!t1).tv := Valitud;
			if !((!t2).tv) = Vaatlemata then (!t2).tv := Valitud; 
		);;

(* funktsioon, mis lisab serva, st mrgib tema ja tema tipud vaadelduks *)
let lisaServ(lisatavServ) = 
	match lisatavServ with
		| {tipp1 = t1; tipp2 = t2; sv = v} -> (
			v := Vaadeldud;
			(!t1).tv := Vaadeldud;
			(!t2).tv := Vaadeldud;
		);;

(* funktsioon listist korduvate elementide eemaldamiseks algusest, tagades jrjekorra silimise lpust vaadates. *)
let rec eemaldaKorduvadAlgusest(list) =
	match list with
		| x::xs -> if List.mem x xs then xs else x :: eemaldaKorduvadAlgusest(xs)
		| [] -> [];;

(* funktsioon listist korduvate elementide eemaldamiseks lpust, tagades jrjekorra silimise algusest vaadates.*)
let eemaldaKorduvadLopust(list) = List.rev (eemaldaKorduvadAlgusest(List.rev list));;

(* funktsioon jrgmiste tippude snena esitamiseks *)
(*let string_of_j2rgmisedTipud(servad) = 
	let tipud =	List.map (fun s -> if !(!(s.tipp1).tv) = Vaatlemata then !(s.tipp1) else !(s.tipp2)) servad in
	let unikaalsedTipud = eemaldaKorduvadLopust(tipud) in
	"Jrgmised tipud: " ^ string_of_tipud(unikaalsedTipud);;*)
	
let string_of_j2rgmisedServad(servad) =
	"Jrgmised servad: " ^ string_of_servad(servad);;

(* funktsioon nimekirjade uuendamiseks *)
let lisatekst() =
	nk1 := string_of_toodeldudTipud(!toodeldudTipud);
	(*nk2 := string_of_j2rgmisedTipud(!j2rgmisedServad);;*)
	nk2 := string_of_j2rgmisedServad(!j2rgmisedServad);;

(* algoritmi algus, mille kigus nullime ra tdeldud tipud (tidetud graafikontrollis) ja jrgmised servad *)
let algus() =
	tekst := "Laiuti lbimise algoritm alustab valitud algtipust.";
	toodeldudTipud := [];
	j2rgmisedServad := [];
	i := EsimeneTipp;;
	
(* esimese tipu vaadelduks mrkimine, jrgmiste servade leidmine ja jrjekorda/magasini lisamine *)
let esimeneTipp(algtipp, servad) =
	algtipp.tv := Vaadeldud;
	toodeldudTipud := (!toodeldudTipud) @ [algtipp];
	let js = leiaJ2rgServad(algtipp, servad) in
	j2rgmisedServad := (!j2rgmisedServad) @ js;
	tekst := "Mrgime esimese tipu klastatuks ja lisame " ^ (if !algo = Laiuti then "jrjekorda" else "magasini") ^ 
		(*" need tipud, kuhu sja tdeldud tipust serv viib.";*)
		" kik selle tipuga seotud servad.";
	lisatekst();
	if List.length !j2rgmisedServad = 0											(* kui jrgmiste servade jrjekord on thi, lhme lpule *)
		then i := Lopp
	else i := ServaValik;;																	(* kui jrgmisi servi leidub, lhme uut serva valima *)

(* jrjekorrast/magasinist jrgmise serva valimine ja koos vastava tipuga valituks mrkimine *)
let servaValik(servad) =
	let lisatavServ = List.hd !j2rgmisedServad in 					(* valime jrjekorra algusest jrgmise serva *)
	j2rgmisedServad := List.tl !j2rgmisedServad;						(* eemaldame selle serva jrgmiste servade jrjekorrast *)
	valiServ(lisatavServ);																	(* mrgime serva ja vastava tipu valituks *)
	tekst := "Valime " ^ (if !algo = Laiuti then "jrjekorrast jrgmise" else "magasinist pealmise") ^
			" serva ja eemaldame selle " ^ (if !algo = Laiuti then "jrjekorrast." else "magasinist.");
	lisatekst();
	if !(!(lisatavServ.tipp1).tv) = Vaadeldud && !(!(lisatavServ.tipp2).tv) = Vaadeldud
		then i := SobimatuServ 
	else i := ServaLisamine;;

(* serva ei saa lisada, sest hendab juba tdeldud tippe *)
let sobimatuServ(servad) =
	let sobimatuServ = List.find (fun s -> !(s.sv) = Valitud) servad in
	sobimatuServ.sv := Sobimatu;
	tekst := "Valitud serv hendab juba tdeldud tippe, nii et htki tippu ei ttle.";
	if List.length !j2rgmisedServad = 0
		then i := Lopp
	else i := ServaValik;;

(* valitud serva ja vastava tipu vaadelduks mrkimine ja jrgmiste servade leidmine ja jrjekorda/magasini lisamine *)
let servaLisamine(tipud, servad) =
	let lisatavServ = List.find (fun s -> !(s.sv) = Valitud) servad in
	let lisatavTipp = leiaLisatavTipp(lisatavServ) in				(* lisatav tipp *)
	toodeldudTipud := (!toodeldudTipud) @ [lisatavTipp];		(* lisame selle klastatud tippude hulka *)
	let js = leiaJ2rgServad(lisatavTipp, servad) in 				(* leiame need servad, kuhu valitud tipust viib *)
	j2rgmisedServad := if !algo = Laiuti then !j2rgmisedServad @ js else js @ !j2rgmisedServad; 						
																													(* lisame need jrgmiste servade jrjekorra lppu / magasini *)
	(*j2rgmisedServad := eemalda(!j2rgmisedServad);*) 					(* eemaldame jrgmiste servade jrjekorrast / magasinist need 
																														servad, mis hendavad klastatud tippe *)
	lisaServ(lisatavServ);																	(* mrgime serva ja tema tipud vaadelduks *)
	tekst := "Mrgime serva ttlemata otstipu tdelduks ja lisame " ^ (if !algo = Laiuti then "jrjekorra lppu" else "magasini") ^ 
		" sja tdeldud tipuga seotud servad.";
	lisatekst();
	if List.length !j2rgmisedServad = 0											(* kui jrgmiste servade jrjekord on thi, lhme lpule *)
		then i := Lopp
	else i := ServaValik;;																	(* kui jrgmisi servi leidub, lhme uut serva valima *)

(* algoritmi lpp *)
let lopp() =
	tekst := "Jrjekord on thi. Laiuti lbimise algoritm lpetab, olles leidnud laiuti lbimise puu.";
	lisatekst();
	AlgoBaas.lopp();;

(* algoritmi samm *)
let samm(algtipp, tipud, servad) = 
	match !i with
		| Algus -> algus();
		| EsimeneTipp -> esimeneTipp(algtipp, servad);
		| ServaValik -> servaValik(servad);
		| SobimatuServ -> sobimatuServ(servad);
		| ServaLisamine -> servaLisamine(tipud, servad);
		| Lopp -> lopp();
		| _ -> ();;

