(* moodul Prim teostab sammsammulist Primi algoritmi lbimngu *) 

open Struktuurid;;
open AlgoBaas;;

let lisatudTipp = ref(tyhiTipp);;		(* lisatud tipp *)

let sej = ref([]);;									(* servade eelistusjrjekord *)

(* funktsioon, mis tagastab, kas serv on sobiv, st kui ta on suunatud, siis ega tema esimene tipp vaadeldud pole, kui aga *)
(* mittesuunatud, siis ega mlemad tipud vaadeldud pole *)
let sobivJ2rgServ tipp serv =
	match serv with 
		| {tipp1 = t1; tipp2 = t2; nool = n} ->
			match n with
				| false -> !t1 = tipp && !((!t2).tv) <> Vaadeldud || !t2 = tipp && !((!t1).tv) <> Vaadeldud	(* mittesuunatud graaf *)
				| true -> !t1 = tipp && !((!t2).tv) <> Vaadeldud;;																					(* suunatud graaf *)

(* funktsioon jrgmiste sobivate servade leidmiseks *)
let leiaJ2rgServad(tipp, servad) =
	List.filter (sobivJ2rgServ tipp) servad;;

(* funktsioon jrjekorras olevate servade snena esitamiseks *)
let string_of_j2rgmisedServad(servad) =
	"Jrjekorras olevad servad: " ^ string_of_servad(servad);;

(* funktsioon, mis mrgib serva ja tema tipud vaadeldavaks, kui need enne vaatlemata olid *)
let vaatleServ(serv) =
	match serv with
		| {tipp1 = t1; tipp2 = t2;} -> (
			if !((!t1).tv) = Vaatlemata then (!t1).tv := Vaadeldav;
			if !((!t2).tv) = Vaatlemata then (!t2).tv := Vaadeldav;
			serv.sv := Vaadeldav;
		);;

(* funktsioon, mis mrgib serva ja tema tipud valituks, kui nad enne vaadeldavad olid, ja eemaldab serva jrjekorrast *)
let valiServ(serv) =
	match serv with
		| {tipp1 = t1; tipp2 = t2;} -> (
			if !((!t1).tv) = Vaadeldav then (!t1).tv := Valitud;
			if !((!t2).tv) = Vaadeldav then (!t2).tv := Valitud;
			serv.sv := Valitud;
		);;

(* funkstioon, mis mrgib tipu vaadelduks ning omistab lisatudTipule *)
let lisaTipp(tipp) =
	tipp.tv := Vaadeldud;
	lisatudTipp := tipp;;

(* algoritmi algus *)
let algus(servad) =
	tekst := "Primi algoritm alustab valitud tipust.";
	(*nk1 := string_of_sej(!sej);*)
	i := EsimeneTipp;;

(* esimese tipu vaadelduks mrkimine *)
let esimeneTipp(algtipp, servad) =
	lisaTipp(algtipp);
	tekst := "Mrgime esimese tipu klastatuks.";
	nk1 := string_of_sej(!sej);
	i := ServaVaatlus;;

(* seda tippu teiste klastamata tippudega hendavate servade leidmine, kaalude phjal sortimine ja vaadeldavateks mrkimine *)
let servaVaatlus(servad) =
	let js = leiaJ2rgServad(!lisatudTipp, servad) in				(* leiame servad, mis viivad sellest tipust klastamata tippu *)
	List.iter vaatleServ js;																(* mrgime need vaadeldavateks *)
	sej := !sej @ js;
	sej := sordiJ2rjekord(!sej);
	tekst := "Lisame eelistusjrjekorda kik servad, mis sja lisatud tippu mne klastamata tipuga hendavad.";
	nk1 := string_of_sej(!sej);
		if List.length !sej = 0																(* kui enam servi lisada ei saa, siis lhme lpule *)								
		then i := Lopp
	else i := ServaValik;;																	(* vastasel juhul lhme jrgmisi servi lisama *)

(* lhima eelistusjrjekorras oleva serva valituks mrkimine *)
let servaValik() =
	let lyhimServ = List.hd !sej in																			(* valime eelistusjrjekorrast lhima serva *)
	valiServ(lyhimServ);																								(* mrgime selle valituks *)
	tekst := "Valime eelistusjrjekorrast vikseima kaaluga serva.";
	nk1 := string_of_sej(!sej);
	i := ServaLisamine;;

(* valitud serva lisamine, kui tema ks otstippe on klastamata, ja eelistusjrjekorrast eemaldamine *)
let servaLisamine(algtipp, tipud, servad) =
	let lisatavServ = List.hd !sej in
	if !(!(lisatavServ.tipp1).tv) = Vaadeldud && !(!(lisatavServ.tipp2).tv) = Vaadeldud
		then (	(* serv hendab juba vaadeldud tippe -> muudame sobimatuks *)
			lisatavServ.sv := Sobimatu;
			tekst := "Seda serva ei lisa, sest see hendab kahte klastatud tippu. Eemaldame serva eelistusjrjekorrast.";
			sej := List.tl !sej;																							(* eemaldame jrjekorrast leliigsed servad *)
			nk1 := string_of_sej(!sej);
			if List.length !sej = 0																							(* kui enam servi lisada ei saa, siis lhme lpule *)
  			then i := Lopp
			else i := ServaValik
		)
	else (		(* sobib, lisame *)
  	let lisatavTipp = List.find (fun t -> !(t.tv) = Valitud) tipud in
  	lisaTipp(lisatavTipp);																							(* mrgime lisatava tipu vaadelduks *)
  	lisatavServ.sv := Vaadeldud;																				(* mrgime lisatava serva vaadelduks *)
		tekst := "Mrgime valitud serva teise otstipu klastatuks ja eemaldame serva eelistusjrjekorrast.";
		sej := List.tl !sej;																								(* eemaldame jrjekorrast leliigsed servad *)
		nk1 := string_of_sej(!sej);
		i := ServaVaatlus);;																						(* vastasel juhul lhme veel servi vaatlema *)

(* algoritmi lpp *)
let lopp() =
	tekst := "Servade eelistusjrjekord on thi. Algoritm lpetab, olles leidnud minimaalse toespuu.";
	nk1 := string_of_sej(!sej);
	AlgoBaas.lopp();;

(* algoritmi samm *)
let samm(algtipp, tipud, servad) = 
	match !i with
		| Algus -> algus(servad)
		| EsimeneTipp -> esimeneTipp(algtipp, servad)
		| ServaVaatlus -> servaVaatlus(servad)
		| ServaValik -> servaValik()
		| ServaLisamine -> servaLisamine(algtipp, tipud, servad)
		| Lopp -> lopp()
		| L2bi -> ()
		| _ -> ();;
		