(* moodul TopoKahn teostab sammsammulist lbimngu topoloogilise jrjestuse leidmiseks Kahni algoritmiga *) 

open Struktuurid;;
open AlgoBaas;;

let sisendastmed = Hashtbl.create 10;; 		(*sisuliselt map tipp.nimi : int *)

let tekkinudJ2rjestus = ref([]);; 				(* siia tekkinud topoloogiline jrjestus *)

(* funktsioon tippude sisendastmete snena esitamiseks *)
let string_of_sisendastmed() =
	let s = Hashtbl.fold (fun k v acc -> k ^ ": " ^ string_of_int(v) ^ ", " ^ acc) sisendastmed "" in
  "Tippude sisendastmed: [" ^ String.sub s 0 (String.length(s) - 2) ^ "]";;

(* funktsioon tippude topoloogilise jrjestuse snena esitamiseks *)
let string_of_topo() =
	"Topoloogiline jrjestus: " ^ string_of_tipud(!tekkinudJ2rjestus);;

(* funktsioon sisendastme uuendamiseks Hashtblis *)
let uuendaSisendastet nr tipp =
	Hashtbl.replace sisendastmed tipp.nimi nr;;

(* funktsioon tipu sisendastme leidmiseks *)
let leiaSisendaste(tipp, servad) =
	List.length (List.filter ((fun t s -> !(s.tipp2) = tipp) tipp) servad);;

(* vhendame sisendtipu sisendastet (v.a. kui see juba 0 on) *)
let v2hendaSisendastet(serv) =
	let t = !(serv.tipp2) in
	let senineSisendaste = Hashtbl.find sisendastmed t.nimi in
	if senineSisendaste > 0
		then uuendaSisendastet (senineSisendaste - 1) t;;

(* valime suvaliselt tipu, millel on sisendaste 0 ning mis pole veel Vaadeldud.*)
let valiTipp(tipud) : Struktuurid.tipp =
	Random.self_init();
	let nsat = List.filter (fun t -> Hashtbl.find sisendastmed t.nimi = 0 && !(t.tv) <> Vaadeldud) tipud in (*nullise sisendastmega tipud *)
	if List.length nsat = 0 
		then failwith("Pole htegi klastamata tippu, mille sisendaste oleks 0. Ei tohiks juhtuda.");
	List.nth nsat (Random.int (List.length nsat));;

(* funktsioon, mis juhul, kui serva lhtetipp on valitud, mrgib serva vaadeldavaks, vhendab tema sisendastet ja *)
(* mrgib vaatlemata sihttipu vaadeldavaks*)
let vaatleServa(serv) =
	match serv with
		| {tipp1 = t1; tipp2 = t2; sv = v} -> (
			if !(!t1.tv) = Valitud
				then (
					v := Vaadeldav;
					v2hendaSisendastet(serv);
					if !(!t2.tv) = Vaatlemata then !t2.tv := Vaadeldav 
				)
		);;
	
(* funktsioon, mis mrgib tipu vaadelduks ja lisab ta tekkinud jrjestuse lppu *)
let lisaTipp(tipp) =
	tipp.tv := Vaadeldud;
	tekkinudJ2rjestus := !tekkinudJ2rjestus @ [tipp];;

(* funktsioon, mis uuendab graafi kohal kuvatavaid nimekirju *)
let lisatekst() =
	nk1 := string_of_sisendastmed();
	nk2 := string_of_topo();;

(* algoritmi algus *)
let algus() =
	Random.self_init();
	tekst := "Kahni algoritm topoloogilise jrjestuse leidmiseks alustab. ";
	i := Vahe1;;

(* paneme iga tipuga vastavusse tema sisendastme *)
let vahe1(tipud, servad) =
	List.iter (fun t -> uuendaSisendastet (leiaSisendaste(t, servad)) t) tipud;
	tekst := "Mrame iga tipuga vastavusse temasse sisenevate servade arvu.";
	nk1 := string_of_sisendastmed();
	i := TipuValik;;

(* suvalise sisendastemga 0 tipu valimine *)
let tipuValik(tipud) =
	tekst := "Valime suvaliselt he tipu, mille sisendaste on 0.";
	lisatekst();
	let valitudTipp = valiTipp(tipud) in
	valitudTipp.tv := Valitud;
	i := ServaVaatlus;;

(* kikide valitud tipust vljuvate servade vaatlemine ja sihttippude sisendastmete vhendamine *)
let servaVaatlus(servad) =
	List.iter vaatleServa servad;
	tekst := "Vaatleme kiki servi, mis valitud tipust vljuvad, ja vhendame sihttippude sisendastet 1 vrra.";
	lisatekst();
	i := ServaLisamine;;

(* valitud tipu lisamine topoloogilisse jrjestusse *)
let servaLisamine(tipud, servad) =
	List.iter (fun t -> if !(t.tv) = Valitud then lisaTipp t) tipud;
	List.iter (fun t -> if !(t.tv) = Vaadeldav then t.tv := Vaatlemata) tipud;
					(* kik teised tipud uuesti Vaadeldav -> Vaatlemata *)
	List.iter (fun s -> if !(s.sv) = Vaadeldav then s.sv := Vaadeldud) servad;
					(* kik Vaadeldavad servad -> Vaadeldud *)
	tekst := "Lisame valitud tipu topoloogilisse jrjestusse.";
	lisatekst();
	if List.for_all (fun t -> !(t.tv) = Vaadeldud) tipud
		then i := Lopp
	else i := TipuValik;;

(* algoritmi lpp *)
let lopp() =
	tekst := "Kik tipud on vaadeldud. Algoritm lpetab, olles leidnud sunatud graafi topoloogilise jrjestuse.";
	lisatekst();
	AlgoBaas.lopp();;

(* algoritmi samm *)
let samm(tipud, servad) = 
	match !i with
		| Algus -> algus()
		| Vahe1 -> vahe1(tipud, servad)
		| TipuValik -> tipuValik(tipud)
		| ServaVaatlus -> servaVaatlus(servad)
		| ServaLisamine -> servaLisamine(tipud, servad)
		| Lopp -> lopp()
		| L2bi -> ()
		| _ -> ();;